Program VOProc;
{ Sorts the polygons in a counter clockwise order }
{$G+,R-}

Uses
  Crt, Bufio;

{$I VERSION.INC}

Type
  FVertex3d = Record
    x, y, z : Real;
  End;
  LCoord = Record
    X, Y, Z : Integer;
  End;
  SPolygon = Record
    a, b, c : Integer;
  End;
  TCoordList = Array[0..5000] of LCoord;
  PCoordList = ^TCoordList;
  TPolygonList = Array[0..5000] of SPolygon;
  PPolygonList = ^TPolygonList;
  TSortList = Array[0..5000] of Boolean;
  PSortList = ^TSortList;
  TType = Array[0..2] of Integer;

Var
  NumSorted : Integer;
  VOName, SaveVOName : String;
  SwapNormals : Boolean;
  AdjacentList : Array[0..15] of Integer;
  Vertex : PCoordList;
  Normals : PCoordList;
  Polygons : PPolygonList;
  VNormals : PCoordList;
  Flag : Word;
  NumVerticies : Word;
  NumPolygons : Word;
  MaxVertInPoly : Byte;
  SortList : PSortList;
  VSortList : PSortList;
  GVOFlag : Array[0..2] of Byte;

Procedure ReCalcNormals;

Var
  Dir1, Dir2 : FVertex3d;
  Normal : FVertex3d;
  Count : Word;
  NormLen : Real;
  RetCode : Integer;

Begin
  For Count := 0 to (NumPolygons - 1) do
    Begin
      Dir1.x := Vertex^[Polygons^[Count].b].x - Vertex^[Polygons^[Count].a].x;
      Dir1.y := Vertex^[Polygons^[Count].b].y - Vertex^[Polygons^[Count].a].y;
      Dir1.z := Vertex^[Polygons^[Count].b].z - Vertex^[Polygons^[Count].a].z;
      Dir2.x := Vertex^[Polygons^[Count].c].x - Vertex^[Polygons^[Count].a].x;
      Dir2.y := Vertex^[Polygons^[Count].c].y - Vertex^[Polygons^[Count].a].y;
      Dir2.z := Vertex^[Polygons^[Count].c].z - Vertex^[Polygons^[Count].a].z;
      Normal.x := (Dir1.y * Dir2.z) - (Dir1.z * Dir2.y);
      Normal.y := (Dir1.z * Dir2.x) - (Dir1.x * Dir2.z);
      Normal.z := (Dir1.x * Dir2.y) - (Dir1.y * Dir2.x);
      NormLen := Sqrt(Sqr(Normal.x) + Sqr(Normal.y) + Sqr(Normal.z));
      If NormLen <> 0
        Then Begin
          Normals^[Count].x := Round((Normal.x / NormLen) * 256);
          Normals^[Count].y := Round((Normal.y / NormLen) * 256);
          Normals^[Count].z := Round((Normal.z / NormLen) * 256);
        End
        Else Begin
          Normals^[Count].x := 0;
          Normals^[Count].y := 0;
          Normals^[Count].z := 0;
        End;
    End;
End;


Procedure LoadVO(Filename : String);

Var
  Diskfile : File;
  Count : Integer;

Begin
  Assign(Diskfile, Filename);
  Reset(Diskfile, 1);
  BlockRead(Diskfile, GVOFlag, 3);
  BlockRead(Diskfile, Flag, 2);
  Flag := 15;
  BlockRead(Diskfile, NumVerticies, 2);
  Getmem(Vertex, NumVerticies*Sizeof(LCoord));
  For Count := 0 to (NumVerticies-1) do
    Begin
      BlockRead(Diskfile, Vertex^[Count].x, 2);
      BlockRead(Diskfile, Vertex^[Count].y, 2);
      BlockRead(Diskfile, Vertex^[Count].z, 2);
    End;
  BlockRead(Diskfile, NumPolygons, 2);
  Getmem(Polygons, NumPolygons*Sizeof(SPolygon));
  For Count := 0 to (NumPolygons-1) do
    Begin
      BlockRead(Diskfile, Polygons^[Count].a, 2);
      BlockRead(Diskfile, Polygons^[Count].b, 2);
      BlockRead(Diskfile, Polygons^[Count].c, 2);
    End;
  Getmem(Normals, NumPolygons*Sizeof(LCoord));
  For Count := 0 to (NumPolygons-1) do
    Begin
      BlockRead(Diskfile, Normals^[Count].x, 2);
      BlockRead(Diskfile, Normals^[Count].y, 2);
      BlockRead(Diskfile, Normals^[Count].z, 2);
    End;
  Getmem(VNormals, NumVerticies*Sizeof(LCoord));
  For Count := 0 to (NumVerticies-1) do
    Begin
      BlockRead(Diskfile, VNormals^[Count].x, 2);
      BlockRead(Diskfile, VNormals^[Count].y, 2);
      BlockRead(Diskfile, VNormals^[Count].z, 2);
    End;
  Close(Diskfile);
  Getmem(SortList, NumPolygons);
  Getmem(VSortList, NumVerticies);
  FillChar(SortList^[0], NumPolygons, Ord(False));
  FillChar(VSortList^[0], NumVerticies, Ord(False));
End;

Procedure IntSwap(Var A, B : Integer);

Var
  Temp : Integer;

Begin
  Temp := A;
  A := B;
  B := Temp;
End;

Function SortAdjPolys(Index1, Index2 : Integer) : Boolean;

Var
  Temp : Integer;
  Match1, Match2, Match3 : Integer;
  Dir : Integer;
  p1, p2 : SPolygon;

Begin
  p1 := Polygons^[Index1];
  p2 := Polygons^[Index2];
  If      (p1.a = p2.a) Then Match1 := 0
  Else If (p1.a = p2.b) Then Match1 := 1
  Else If (p1.a = p2.c) Then Match1 := 2
  Else Match1 := -1;

  If      (p1.b = p2.a) Then Match2 := 0
  Else If (p1.b = p2.b) Then Match2 := 1
  Else If (p1.b = p2.c) Then Match2 := 2
  Else Match2 := -1;

  If      (p1.c = p2.a) Then Match3 := 0
  Else If (p1.c = p2.b) Then Match3 := 1
  Else If (p1.c = p2.c) Then Match3 := 2
  Else Match3 := -1;

  Temp := 0;
  If (Match1 > -1) Then Inc(Temp);
  If (Match2 > -1) Then Inc(Temp);
  If (Match3 > -1) Then Inc(Temp);

  If (Temp <> 2)
    Then Begin
      SortAdjPolys := False;
      Exit;
    End;

  Dir := -1;
  If (Match1 = -1) Then If (((Match2+1) Mod 3) = Match3) Then Dir := 1;
  If (Match2 = -1) Then If (((Match3+1) Mod 3) = Match1) Then Dir := 1;
  If (Match3 = -1) Then If (((Match1+1) Mod 3) = Match2) Then Dir := 1;

  If Dir > 0
    Then Begin
      IntSwap(Polygons^[Index2].b, Polygons^[Index2].c);
      Normals^[Index2].x := -Normals^[Index2].x;
      Normals^[Index2].y := -Normals^[Index2].y;
      Normals^[Index2].z := -Normals^[Index2].z;
    End;

  SortAdjPolys := True;
End;

Procedure BranchSort(BaseIndex : Integer);

Var
  Count : Integer;
  AdjPoly : Integer;

Begin
  For Count := 0 to (NumPolygons-1) do
    Begin
      If Not(SortList^[Count])
        Then Begin
          SortList^[Count] := SortAdjPolys(BaseIndex, Count);
          If SortList^[Count]
            Then Begin
              Inc(NumSorted);
              VSortList^[Polygons^[Count].a] := True;
              VSortList^[Polygons^[Count].b] := True;
              VSortList^[Polygons^[Count].c] := True;
            End;
        End;
    End;
End;

Procedure CCWSort;

Const
  Turn : String = '////||||\\\\----';

Var
  Distance : Real;
  MaxIndex : Integer;
  MaxDistance : Real;
  Count : Integer;
  Dot1, Dot2, Dot3 : Real;
  CCW : Boolean;
  CurrentIndex : Integer;
  TurnCount : Integer;
  i : integer;
  PolyList : Array[0..31] of Integer;
  LongVector : FVertex3d;
  NormalVector : FVertex3d;
  tx, ty, tz : Real;
  AdjHalt : Integer;
  LastCount : Integer;

Begin
  LastCount := -1;
  AdjHalt := 0;
  MaxDistance := 0;
  For i := 0 to (NumVerticies - 1) do
    Begin
      Dot1 := Vertex^[i].x;
      Dot2 := Vertex^[i].y;
      Dot3 := Vertex^[i].z;
      Distance := Sqrt(Sqr(Dot1) + Sqr(Dot2) + Sqr(Dot3));
      If (Distance > MaxDistance) And (VSortList^[i] = False)
        Then Begin
          LongVector.x := Dot1;
          LongVector.y := Dot2;
          LongVector.z := Dot3;
          MaxDistance := Distance;
          MaxIndex := i;
        End;
    End;

  Count := 0;
  For i := 0 to (NumPolygons - 1) do
    Begin
      If ((Polygons^[i].a = MaxIndex) or (Polygons^[i].b = MaxIndex)) or (Polygons^[i].c = MaxIndex)
        Then Begin
          If (SortList^[i] = False)
            Then Begin
              PolyList[Count] := i;
              Inc(Count);
            End;
        End;
    End;

  MaxDistance := 0;
  For i := 0 to (Count - 1) do
    Begin
      tx := Vertex^[Polygons^[PolyList[i]].a].x;
      ty := Vertex^[Polygons^[PolyList[i]].a].y;
      tz := Vertex^[Polygons^[PolyList[i]].a].z;
      Dot1 := Sqrt(Sqr(tx) + Sqr(ty) + Sqr(tz));

      tx := Vertex^[Polygons^[PolyList[i]].b].x;
      ty := Vertex^[Polygons^[PolyList[i]].b].y;
      tz := Vertex^[Polygons^[PolyList[i]].b].z;
      Dot2 := Sqrt(Sqr(tx) + Sqr(ty) + Sqr(tz));

      tx := Vertex^[Polygons^[PolyList[i]].c].x;
      ty := Vertex^[Polygons^[PolyList[i]].c].y;
      tz := Vertex^[Polygons^[PolyList[i]].c].z;
      Dot3 := Sqrt(Sqr(tx) + Sqr(ty) + Sqr(tz));

      Distance := (Dot1 + Dot2 + Dot3);
      If (Distance > MaxDistance)
        Then Begin
          MaxIndex := PolyList[i];
          MaxDistance := Distance;
        End;
    End;

  NormalVector.x := Normals^[MaxIndex].x / 256;
  NormalVector.y := Normals^[MaxIndex].y / 256;
  NormalVector.z := Normals^[MaxIndex].z / 256;

  If (LongVector.x * NormalVector.x + LongVector.y * NormalVector.y + LongVector.z * NormalVector.z) < 0
    Then Begin
      IntSwap(Polygons^[MaxIndex].b, Polygons^[MaxIndex].c);
      ReCalcNormals;
    End;

  SortList^[MaxIndex] := True;
  VSortList^[Polygons^[MaxIndex].a] := True;
  VSortList^[Polygons^[MaxIndex].b] := True;
  VSortList^[Polygons^[MaxIndex].c] := True;
  TurnCount := 0;
  Count := MaxIndex;
  While (NumSorted < NumPolygons) and (AdjHalt < 2*NumPolygons) do
    Begin
      If SortList^[Count]
        Then Begin
          GotoXY(1, WhereY);
          Write('[', Turn[TurnCount+1], ']   [', NumSorted+1, '/', NumPolygons, ']');
          ClrEol;
          LastCount := NumSorted;
          BranchSort(Count);
          TurnCount := (TurnCount + 1) And 15;
        End;
      Count := Count + 1;
      If (Count >= NumPolygons) Then Count := 0;
      If (LastCount = NumSorted)
        Then Inc(AdjHalt)
        Else AdjHalt := 0;
    End;
End;

Procedure ChangeEm;

Var
  Count : Integer;

Begin
  For Count := 0 to (NumPolygons-1) do
    Begin
      IntSwap(Polygons^[Count].b, Polygons^[Count].c);
      Normals^[Count].x := -Normals^[Count].x;
      Normals^[Count].y := -Normals^[Count].y;
      Normals^[Count].z := -Normals^[Count].z;
    End;
End;

Procedure CalcVertexNormals;

Var
  RetCode : Integer;
  i, j : Integer;
  Xa, Ya, Za : Longint;
  Len : Real;

Begin
  RetCode := 0;
  For i := 0 to (NumVerticies - 1) do
    Begin
      Xa := 0;
      Ya := 0;
      Za := 0;
      For j := 0 to (NumPolygons - 1) do
        Begin
          If ((Polygons^[j].a = i) or (Polygons^[j].b = i)) or
             (Polygons^[j].c = i)
          Then Begin
            Inc(Xa, Normals^[j].x);
            Inc(Ya, Normals^[j].y);
            Inc(Za, Normals^[j].z);
          End;
        End;
      Len := Sqrt(Sqr(Xa) + Sqr(Ya) + Sqr(Za));
      If Len = 0
        Then Begin
          RetCode := 6;
        End
        Else Begin
          VNormals^[i].x := Round((Xa/Len)*256);
          VNormals^[i].y := Round((Ya/Len)*256);
          VNormals^[i].z := Round((Za/Len)*256);
        End;
    End;
End;

Procedure SaveVO(FileName : String);

Var
  Diskfile : bFile;
  Count : Integer;
  FileBuffer : Array[0..1023] of Byte;

Begin
  FAssign(Diskfile, Filename);
  FSetBuffer(Diskfile, FileBuffer, Sizeof(FileBuffer));
  FReWrite(Diskfile);
  FWrite(Diskfile, GVOFlag, 3);
  FWrite(Diskfile, Flag, 2);
  FWrite(Diskfile, NumVerticies, 2);
  For Count := 0 to (NumVerticies-1) do
    Begin
      FWrite(Diskfile, Vertex^[Count].x, 2);
      FWrite(Diskfile, Vertex^[Count].y, 2);
      FWrite(Diskfile, Vertex^[Count].z, 2);
    End;
  FWrite(Diskfile, NumPolygons, 2);
  For Count := 0 to (NumPolygons-1) do
    Begin
      FWrite(Diskfile, Polygons^[Count].a, 2);
      FWrite(Diskfile, Polygons^[Count].b, 2);
      FWrite(Diskfile, Polygons^[Count].c, 2);
    End;
  For Count := 0 to (NumPolygons-1) do
    Begin
      FWrite(Diskfile, Normals^[Count].x, 2);
      FWrite(Diskfile, Normals^[Count].y, 2);
      FWrite(Diskfile, Normals^[Count].z, 2);
    End;
  For Count := 0 to (NumVerticies-1) do
    Begin
      FWrite(Diskfile, VNormals^[Count].x, 2);
      FWrite(Diskfile, VNormals^[Count].y, 2);
      FWrite(Diskfile, VNormals^[Count].z, 2);
    End;
  Flush(Diskfile);
  FClose(Diskfile);
End;

Function UpStr(S : String) : String;

Var
  Count : Integer;
  t : String;

Begin
  t := '';
  For Count := 1 to Length(S) do
    t := t + UpCase(S[Count]);
  UpStr := t;
End;

Procedure Help;

Begin
  Writeln('Usage:   CCWS <infile> [<outfile>] [-s]');
  Writeln('  infile - name of input VO file');
  Writeln('  outfile - optional output filename');
  Writeln('  s - swap direction of normals');
  Halt(0);
End;

Procedure Parse;

Var
  Temp : String;
  InFlag : Boolean;
  OutFlag : Boolean;
  C : Integer;

Begin
  Writeln('Counter Clockwise Sorting Utility v', VERSION_MAJOR, '.',VERSION_MINOR,'                  Alex Chalfin 11/5/95');
  Writeln;
  InFlag := False;
  OutFlag := False;
  SwapNormals := False;
  If ParamCount = 0
    Then Help;
  For C := 1 to ParamCount do
    Begin
      Temp := UpStr(ParamStr(C));
      If (Temp[1] = '-') or (Temp[1] = '/')
        Then Begin
          If Temp[2] = 'S'
            Then SwapNormals := True;
          If Temp[2] = '?'
            Then Help;
          If Temp[2] = 'H'
            Then Help;
        End
        Else Begin
          If Not(InFlag)
            Then Begin
              VOName := Temp;
              InFlag := True;
            End
            Else Begin
              SaveVOName:= Temp;
              OutFlag := True;
            End;
        End;
    End;
  If Not(InFlag)
    Then Begin
      Writeln('No input file specified!');
      Halt(0);
    End;
  If Pos('.', VOName) = 0
    Then VOName := VOName + '.GVO';
  If Not(OutFlag)
    Then Begin
      SaveVOName := VOName;
      While(SaveVOName[Length(SaveVOName)] <> '.') do
        Delete(SaveVOName, Length(SaveVOName), 1);
      SaveVOName := SaveVOName + 'GVO';
    End
    Else Begin
      If Pos('.', SaveVOName) = 0
        Then SaveVOName := SaveVOName + '.GVO';
    End;
End;

Begin
  Parse;
  LoadVO(VOName);
  ReCalcNormals;
  If SwapNormals
    Then Begin
      Writeln('Swapping normals.');
      ChangeEm;
    End
    Else Begin
      NumSorted := 1;
      While (NumSorted < (NumPolygons-1)) and Not(KeyPressed) do CCWSort;
    End;
  CalcVertexNormals;
  SaveVO(SaveVOName);
  Writeln;
End.
