Unit PCX;
{$I-}

Interface

Type
  RGB = Record
    r,g,b : Byte;
  End;
  Palette = Array[0..255] of RGB;

Function LoadPCXFile(Filename : String; Var BitMap : Pointer; Var Pal : Palette; Var Size : Word; FOfs : Longint) : Integer;
Procedure SavePCXFile(Filename : String; XSize, YSize : Word; Screen : Pointer; Pal : Palette);

Implementation

Uses BufIO;

Type
  PCXHeaderType = Record
    Manufacturer : Byte;
    Version      : Byte;
    Encoding     : Byte;
    BitsPerPixel : Byte;
    XMin, YMin, XMax, YMax : Integer;
    HDpi, VDpi   : Integer;                { 0C, 0E }
    ColorMap     : Array[0..47] of Byte;   { 10 - 56 }
    Reserved     : Byte;                   { 57 }
    NPlanes      : Byte;
    BytesPerLine : Integer;
    PaletteInfo  : Integer;
    HScreenSize  : Integer;
    VScreenSize  : Integer;
    Filler       : Array[0..53] of Byte;
  End;


Procedure SavePCXFile(Filename : String; XSize, YSize : Word; Screen : Pointer; Pal : Palette);

Var
  Header : PCXHeaderType;
  RunLen : Byte;
  Value : Byte;
  TempB : Byte;
  FileBuffer : Array[0..1023] of Byte;
  Diskfile : bFile;
  SSeg, SOfs : Word;
  ScrOfs : Word;

Begin
  Header.Manufacturer := 10;
  Header.Version := 5;
  Header.Encoding := 1;
  Header.BitsPerPixel := 8;
  Header.XMin := 0;
  Header.YMin := 0;
  Header.XMax := XSize - 1;
  Header.YMax := YSize - 1;
  Header.HDpi := 320;
  Header.VDpi := 200;
  Header.Reserved := 0;
  Header.NPlanes := 1;
  Header.BytesPerLine := XSize;
  Header.PaletteInfo := 1;
  Header.HScreenSize := XSize;
  Header.VScreenSize := YSize;
  FillChar(Header.ColorMap, Sizeof(Header.ColorMap), 0);
  FillChar(Header.Filler, Sizeof(Header.Filler), 0);

  SSeg := Seg(Screen^);
  SOfs := Ofs(Screen^);

  FAssign(Diskfile, Filename);
  FSetBuffer(Diskfile, FileBuffer, Sizeof(FileBuffer));
  FReWrite(Diskfile);
  FWrite(Diskfile, Header, Sizeof(Header));

  ScrOfs := 0;
  While (ScrOfs < XSize*YSize) do
    Begin
      RunLen := 1;
      Value := Mem[SSeg:SOfs];
      While ((Mem[SSeg:SOfs + RunLen] = Value) And (RunLen < 63)) And (ScrOfs < XSize*YSize) do
        Inc(RunLen);
      If (RunLen > 1)
        Then Begin
          TempB := RunLen Or $C0;
          FWrite(Diskfile, TempB, 1);
          FWrite(Diskfile, Value, 1);
        End
        Else Begin
          If Value >= $C0
            Then Begin
              TempB := $C1;
              FWrite(Diskfile, TempB, 1);
            End;
          FWrite(Diskfile, Value, 1);
        End;
      Inc(ScrOfs, RunLen);
      Inc(SOfs, RunLen);
    End;

  TempB := $0C;
  FWrite(Diskfile, TempB, 1);
  FWrite(Diskfile, Pal, Sizeof(Pal));
  Flush(Diskfile);
  FClose(Diskfile);
End;

Function LoadPCXFile(Filename : String; Var BitMap : Pointer; Var Pal : Palette; Var Size : Word; FOfs : Longint) : Integer;
{ Returns:  0 - Operation successful    }
{           1 - File not found          }
{           2 - Error reading from file }
{           3 - Not enough memory       }

Var
  Header : PCXHeaderType;
  TempPtr : Pointer;
  pSize : Longint;
  TextureOffset : Longint;
  RunLen : Byte;
  Value : Byte;
  InRead : Word;
  Buffer : Array[0..1023] of Byte;
  Diskfile : bFile;

Begin
  FAssign(Diskfile, FileName);
  FSetBuffer(Diskfile, Buffer, Sizeof(Buffer));
  FReset(Diskfile);
  If IOResult <> 0
    Then Begin
      LoadPCXFile := 1;
      Exit;
    End;
  FSeek(Diskfile, FOfs);
  InRead := FRead(Diskfile, Header, Sizeof(Header));
  If IOResult <> 0
    Then Begin
      FClose(Diskfile);
      LoadPCXFile := 2;
      Exit;
    End;
  pSize := Longint(Header.XMax-Header.XMin+1)*Longint(Header.yMax-Header.YMin+1);
  If (pSize > 65536) and (pSize < MaxAvail)
    Then Begin
      FClose(Diskfile);
      LoadPCXFile := 3;
      Exit;
    End;
  TempPtr := Nil;
  Getmem(TempPtr, pSize-1);
  TextureOffset := 0;
  While TextureOffset < pSize do
    Begin
      InRead := FRead(Diskfile, RunLen, 1);
      If IOResult <> 0
        Then Begin
          FClose(Diskfile);
          Freemem(TempPtr, Size-1);
          LoadPCXFile := 2;
          Exit;
        End;
      If (RunLen and $C0) = $C0
        Then Begin
          RunLen := RunLen And $3f;
          InRead := FRead(Diskfile, Value, 1);
          If IOResult <> 0
            Then Begin
              FClose(Diskfile);
              Freemem(TempPtr, Size-1);
              LoadPCXFile := 2;
              Exit;
            End;
        End
        Else Begin
          Value := RunLen;
          RunLen := 1;
        End;
      While (RunLen >= 1) and (TextureOffset < pSize) do
        Begin
          Mem[Seg(TempPtr^):Ofs(TempPtr^)+TextureOffset] := Value;
          TextureOffset := TextureOffset + 1;
          RunLen := RunLen - 1;
        End;
    End;

  Repeat
    InRead := FRead(Diskfile, Value, 1);
    If IOResult <> 0
      Then Begin
        FClose(Diskfile);
        Freemem(TempPtr, Size-1);
        LoadPCXFile := 2;
        Exit;
      End;
  Until (Value = 12);
  InRead := FRead(Diskfile, Pal, Sizeof(Palette));
  If IOResult <> 0
    Then Begin
      FClose(Diskfile);
      Freemem(TempPtr, Size-1);
      LoadPCXFile := 2;
      Exit;
    End;
  FClose(Diskfile);

  For Value := 0 to 255 do
    Begin
      Pal[Value].r := Pal[Value].r Div 4;
      Pal[Value].g := Pal[Value].g Div 4;
      Pal[Value].b := Pal[Value].b Div 4;
    End;
  Size := pSize;
  BitMap := TempPtr;
  LoadPCXFile := 0;
End;

End.